import random
from math2 import Vector, Segment
from render import FloorQuad, WallQuad

WALL_N = 1
WALL_E = 2
WALL_S = 4
WALL_W = 8
WALL_ALL = 15
VISITED = 16

class Backtracker(object):
	"""Generates classic twisted mazes.

	http://en.wikipedia.org/wiki/Maze_generation_algorithm#Recursive_backtracker
	"""
	def __init__(self, w, h, ppasses=5):
		self.w = w
		self.h = h
		self.data = [[ WALL_ALL ] * h for x in range(w)]

		self.bw = bw = w * 2 + 1
		self.bh = bh = h * 2 + 1
		self.blocks = [[ 0 ] * bh for x in range(bw)]

		self.ppasses = ppasses

	neighbors = [
		(0, 1, WALL_S), 
		(0, -1, WALL_N), 
		(1, 0, WALL_E), 
		(-1, 0, WALL_W)
	]

	def generate_corridors(self, callback=None):
		data = self.data
		neighbors = self.neighbors

		breakwall = {
			WALL_N: WALL_S,
			WALL_S: WALL_N,
			WALL_E: WALL_W,
			WALL_W: WALL_E
		}

		w = self.w
		h = self.h

		cx, cy = 0, 0
		stack = [(cx, cy)]

		while stack:

			data[cx][cy] |= VISITED
			
			neigh = []
			for nx, ny, wall in neighbors:
				nx += cx
				ny += cy


				if not (0 <= nx < w and 0 <= ny < h):
					continue

				if not (data[nx][ny] & VISITED):
					neigh.append((nx, ny, wall))

			if neigh:
				nx, ny, wall = random.choice(neigh)
				stack.append((cx, cy))
				
				data[cx][cy] &= ~wall
				data[nx][ny] &= ~breakwall[wall]

				cx, cy = nx, ny
			else:
				cx, cy = stack.pop()

	def generate_rooms(self):
		for room in range(random.randint(4, 10)):
			x = random.randint(0, self.w - 10)
			y = random.randint(0, self.h - 10)
			w = random.randint(3, 8)
			h = random.randint(3, 8)

			for cy in range(y, y+h):
				for cx in range(x, x+w):
					self.data[cx][cy] = 0

	def generate(self):
		self.generate_corridors()
		self.generate_loops()
	
	def generate_loops(self):
		for item in range(random.randint(15, 20)):
			rx = random.randint(2, self.w - 2)
			ry = random.randint(2, self.h - 2)
			self.data[rx][ry] = 0
	
	def iterate(self):
		for y in range(self.h):
			for x in range(self.w):
				yield (x, y, self.data[x][y])

	#offsets = [(-1, -1), (0, -1), (1, -1), (-1, 0), (1, 0), (-1, 1), (0, 1), (1, 1)]
	offsets = [(-1, 0), (0, -1), (1, 0), (0, 1)]
	def get_neighbors(self, x, y):
		values = []
		for dx, dy in self.offsets:
			if self.bget_at(x + dx, y + dy):
				values.append((dx, dy))
	
		return values

	def iterate_blocks(self, value=1):
		for x, row in enumerate(self.blocks):
			for y, v in enumerate(row):
				if v == value:
					yield (x, y)

	def bset_at(self, x, y, v):
		if 0 <= x < self.bw and 0 <= y < self.bh:
			self.blocks[x][y] = v

	def bget_at(self, x, y, oob=0):
		if 0 <= x < self.bw and 0 <= y < self.bh:
			return self.blocks[x][y]
		else:
			return oob

	def build_blocks(self):
		# Convert from bitmap into block arrays.
		for y in range(0, self.bh, 2):
			for x in range(0, self.bw, 2):
				self.bset_at(x, y, 1)

		neighbors = self.neighbors
		for x, y, v in self.iterate():
			x = x * 2 + 1
			y = y * 2 + 1
			
			for dx, dy, bit in neighbors:
				if v & bit:
					self.bset_at(x + dx, y + dy, 1)
			
		# Postprocess.
		for dummy in range(self.ppasses):
			for x, y in self.iterate_blocks():
				neighbors = len(self.get_neighbors(x, y))
				# Lonely blocks are removed.
				if neighbors < 2:
					self.bset_at(x, y, 0)
				# Surrounded blocks are filled.
				if neighbors > 3:
					self.bset_at(x, y, 1)

	def build(self, walltex, floortex, scale=None):
		segments = []
		shapes = []
		self.build_blocks()

		neighbors = self.neighbors
		for x, y in self.iterate_blocks():
			for dx, dy, bit in neighbors:
				if self.bget_at(x + dx, y + dy, 1):
					continue
				seg, shp = self.build_segment(x, y, dx, dy, walltex)
				segments.append(seg)
				shapes.append(shp)

		unit = Vector(0.25, 0.25)
		for x, y in self.iterate_blocks(0):
			loc = Vector(x, y).mul(0.5)
			p1 = loc.subv(unit)
			p2 = p1.addi(0.5, 0)
			p3 = p2.addi(0, 0.5)
			p4 = p3.addi(-0.5, 0)
			shapes.append(FloorQuad(floortex, [p1, p2, p3, p4]))
		
		return segments, shapes
	
	def build_segment(self, x, y, dx, dy, tex):
		loc = Vector(x, y)
		head = Vector(dx, dy).normal()
		unit = head.mul(0.25)
		su = head.mul(0.2)

		n1 = self.bget_at(*loc.addv(head.right()).tupi())
		n2 = self.bget_at(*loc.addv(head.left()).tupi())
		n1n = self.bget_at(*loc.addv(head.right()).addv(head).tupi())
		n2n = self.bget_at(*loc.addv(head.left()).addv(head).tupi())

		loc = loc.mul(0.5)

		p1 = loc.addv(unit).addv(unit.right())
		p2 = loc.addv(unit).addv(unit.left())

		p1o = p1.subv(su)
		p2o = p2.subv(su)

		if n1 and n1n:
			p1o = p1o.addv(su.right())
		elif not (n1 or n1n):
			p1o = p1o.subv(su.right())
		if n2 and n2n:
			p2o = p2o.addv(su.left())
		elif not (n2 or n2n):
			p2o = p2o.subv(su.left())

		seg = Segment(p1, p2)
		seg.quad = WallQuad(tex, [p1, p2, p2o, p1o])

		return seg, seg.quad

if __name__ == '__main__':
	import pygame
	from pygame.locals import *
	from OpenGL.GL import *
	from render import Texture
	
	disp = pygame.display.set_mode((600, 600), DOUBLEBUF | OPENGL)
	clock = pygame.time.Clock()
	
	wtex = Texture('data/wall-test.png')
	ftex = Texture('data/floor-test.png')

	wtex.load()
	ftex.load()

	glEnable(GL_TEXTURE_2D)

	maze = Backtracker(10, 10, 2)
	maze.generate()
	segments, shapes = maze.build(wtex, ftex)
	blocks = maze.blocks

	#white = (255, 255, 255)

	#d = 25

	#d *= 2
	#linemaze = pygame.Surface((600, 600))
	#for x, y, v in maze.iterate():
	#	if v & WALL_N:
	#		pygame.draw.line(linemaze, white, (x*d, y*d), ((x+1)*d, y*d))
	#	if v & WALL_S:
	#		pygame.draw.line(linemaze, white, (x*d, (y+1)*d), ((x+1)*d, (y+1)*d))
	#	if v & WALL_W:
	#		pygame.draw.line(linemaze, white, (x*d, y*d), (x*d, (y+1)*d))
	#	if v & WALL_E:
	#		pygame.draw.line(linemaze, white, ((x+1)*d, y*d), ((x+1)*d, (y+1)*d))

	#d /= 2
	#blockmaze = pygame.Surface((600, 600))
	#for x, col in enumerate(blocks):
	#	for y, v in enumerate(col):
	#		if v:
	#			pygame.draw.rect(blockmaze, white, (x*d, y*d, d, d))
	#
	#segmaze = pygame.Surface((600, 600))
	#for seg in segments:
	#	p1 = seg.p1.mul(d).tupi()
	#	p2 = seg.p2.mul(d).tupi()
	#	pygame.draw.line(segmaze, white, p1, p2)
	#for poly in shapes:
	#	points = poly.points[:]
	#	points.append(points[0])

	#	for p1, p2 in zip(points[:-1], points[1:]):
	#		p1 = p1.mul(d).tupi()
	#		p2 = p2.mul(d).tupi()
	#		pygame.draw.line(segmaze, (0, 255, 255), p1, p2)
	#
	#showsurf = linemaze
	glLoadIdentity()
	glTranslate(-1, -1, 0)
	glScalef(0.2, 0.2, 1)

	while 1:
		for ev in pygame.event.get():
			if ev.type == KEYDOWN:
				if ev.key == K_ESCAPE:
					break
				#if ev.key == K_1:
				#	showsurf = linemaze
				#if ev.key == K_2:
				#	showsurf = blockmaze
				#if ev.key == K_3:
				#	showsurf = segmaze

		glClear(GL_COLOR_BUFFER_BIT)

		for poly in shapes:
			poly.render()

		#disp.blit(showsurf, (0, 0))			
		#pygame.display.update()
		pygame.display.flip()
		clock.tick(50)
