"""Combine different mazes into a single world."""

from math2 import Shape, Segment, Vector, Matrix
from random import choice
from render import FloorQuad, WallQuad, VisibleShape

def combine_shapes(s1, s2, head, floor):
	s1_c, s1_d = s1.center_and_size()
	s2_c, s2_d = s2.center_and_size()

	# Get a segment from of bbox in the direction head.
	side = s1_c.addv(s1_d.scale(head.mul(0.5)))
	p1 = side.subv(s1_d.scale(head.left()))
	p2 = side.subv(s1_d.scale(head.right()))
	s1_cseg = Shape([Segment(p1, p2)])

	side = s2_c.addv(s2_d.scale(head.mul(-0.5)))
	p1 = side.subv(s2_d.scale(head.left()))
	p2 = side.subv(s2_d.scale(head.right()))
	s2_cseg = Shape([Segment(p1, p2)])

	# Walls that we will connect.
	s1_walls = [w for i, w in s1_cseg.contacts(s1) 
			if isinstance(i, Segment)]
	s2_walls = [w for i, w in s2_cseg.contacts(s2) 
			if isinstance(i, Segment)]

	assert(s1_walls)
	assert(s2_walls)

	# Pick two random segments from both shapes.
	s1_dock = choice(s1_walls)
	s2_dock = choice(s2_walls)

	s1_base = s1_dock.middle()
	s2_base = s2_dock.middle()

	# Get amount by which we need to move s2.
	offset = s1_base.subv(s2_base)

	# Add connection margin.
	offset = offset.addv(head.mul(3))

	# # Remove docks.
	# s1.remove(s1_dock)
	# s2.remove(s2_dock)

	# # Remove dock's quads.
	# s1.remove_quad(s1_dock.quad)
	# s2.remove_quad(s2_dock.quad)

	# Move second shape to dock position.
	s2_segs = s2.real_segments[:]
	s2_quads = s2.real_quads[:]
	s2_segs.remove(s2_dock)
	s2_quads.remove(s2_dock.quad)

	s2 = VisibleShape(s2_segs, s2_quads)
	s2.move(offset, Matrix.mident())
	s2_dock = s2_dock.transform(offset, Matrix.mident())

	# Make dock walls.
	if s1_dock.length() > s2_dock.length():
		# We need to run corridor into middle of the s1_dock.
		dock_s, dock_q, fixes = \
			make_split_corridor(s1, s1_dock, s2, s2_dock, floor)
	elif s1_dock.length() < s2_dock.length():
		# We need to run corridor into middle of the s2_dock.
		dock_s, dock_q, fixes = \
			make_split_corridor(s2, s2_dock, s1, s1_dock, floor)
	else:
		# They are the same, just connect endpoints.
		dock_s, dock_q, fixes = \
			make_corridor(s1, s1_dock, s2, s2_dock, floor)

	new_segments = s1.real_segments + s2.real_segments + dock_s
	new_quads = s1.real_quads + s2.real_quads + dock_q

	new_segments.remove(s1_dock)
	new_quads.remove(s1_dock.quad)

	print fixes

	for seg, old_quad, new_quad in fixes:
		new_seg = Segment(seg.p1, seg.p2)
		new_seg.quad = new_quad

		new_segments.remove(seg)
		new_quads.remove(old_quad)

		new_segments.append(new_seg)
		new_quads.append(new_quad)

	return VisibleShape(new_segments, new_quads)
		
def make_corridor(s1, s1_dock, s2, s2_dock, floortex):
	w1 = Segment(s1_dock.p1, s2_dock.p2)
	w2 = Segment(s2_dock.p2, s1_dock.p1)

	u1 = w1.p2.subv(w1.p1).normal().mul(0.2)
	u2 = w2.p2.subv(w2.p1).normal().mul(0.2)

	a1 = w1.p1.addv(u1.left()).addv(u1)
	b1 = w1.p2.addv(u1.left()).subv(u1)

	a2 = w2.p1.addv(u2.left()).addv(u2)
	b2 = w2.p2.addv(u2.left()).subv(u2)

	tex = s1_dock.quad.texture

	w1.quad = WallQuad(tex, [w1.p1, w1.p2, b1, a1])
	w2.quad = WallQuad(tex, [w2.p1, w2.p2, b2, a2])

	floor = FloorQuad(floortex, [w1.p1, w2.p2, w1.p1, w2.p2])

	#fixed1 = fix_docking_quads(s2, s2_dock, a1, b2)
	#fixed2 = fix_docking_quads(s1, s1_dock, b1, a2)

	return [w1, w2], [w1.quad, w2.quad, floor], [] # fixed1 + fixed2

def make_split_corridor(s1, s1_dock, s2, s2_dock, floortex):
	"""s1_dock is longer than s2_dock, so must be split."""

	p1, pu1 = s1_dock.project(s2_dock.p1)
	p2, pu2 = s1_dock.project(s2_dock.p2)

	w1 = Segment(s2_dock.p1, p1)
	w2 = Segment(p2, s2_dock.p2)

	w3 = Segment(s1_dock.p1, p2)
	w4 = Segment(p1, s1_dock.p2)

	# 0.2 is global wall quad thickness. Must be a named but time, TIME!
	u1 = w1.p2.subv(w1.p1).normal().mul(0.2)
	u2 = w2.p2.subv(w2.p1).normal().mul(0.2)
	
	a1 = w1.p1.addv(u1.left()).addv(u1)
	b1 = w1.p2.addv(u1.left()).subv(u1)

	a2 = w2.p1.addv(u2.left()).addv(u2)
	b2 = w2.p2.addv(u2.left()).subv(u2)

	tex = s1_dock.quad.texture

	w1.quad = WallQuad(tex, [w1.p1, w1.p2, b1, a1])
	w2.quad = WallQuad(tex, [w2.p1, w2.p2, b2, a2])

	q1 = s1_dock.quad.vertices
	w3.quad = WallQuad(tex, [w3.p1, w3.p2, a2, q1[3]])
	w4.quad = WallQuad(tex, [w4.p1, w4.p2, q1[2], b1])

	floor = FloorQuad(floortex, [w1.p1, w1.p2, w2.p1, w2.p2])

	# Find s2_dock neighbors, and adjust their quads to make them nicier.
	#fixed = fix_docking_quads(s2, s2_dock, a1, b2)
	fixed = []

	segments = [w1, w2, w3, w4]
	quads = [w1.quad, w2.quad, w3.quad, w4.quad, floor]

	return segments, quads, fixed

def fix_docking_quads(s2, s2_dock, a1, b2):
	n1 = None
	n2 = None

	for seg in s2.real_segments:
		if seg.p2 == s2_dock.p1:
			n1 = seg
			n1v = n1.quad.vertices
		if seg.p1 == s2_dock.p2:
			n2 = seg
			n2v = n2.quad.vertices

	segments = []
	old_quads = []
	new_quads = []

	if n1 and n1v[2] != a1:
		segments.append(n1)
		old_quads.append(n1.quad)
		new_quads.append(WallQuad(s2_dock.quad.texture, 
				[n1v[0], n1v[1], a1, n1v[3]]))

	if n2 and n2v[3] != b2:
		segments.append(n2)
		old_quads.append(n2.quad)
		new_quads.append(WallQuad(s2_dock.quad.texture,
				[n2v[0], n2v[1], n2v[2], b2]))
	
	return zip(segments, old_quads, new_quads)

if __name__ == '__main__':
	import pygame
	from pygame.locals import *
	from OpenGL.GL import *
	from backtracker import Backtracker
	from digger import Digger
	from render import Texture

	pygame.init()
	disp = pygame.display.set_mode((600, 600), DOUBLEBUF | OPENGL)

	wtex = Texture('wall-test.png')
	ftex = Texture('floor-test.png')
	wtex.load()
	ftex.load()

	def make_shape():
		mazes = [
			Digger(),
			Backtracker(30, 30, 15)
		]

		print 'Designing world.'
		for maze in mazes:
			maze.generate()

		print 'Building world.'
		shapes = [VisibleShape(*maze.build(wtex, ftex)) for maze in mazes]

		print 'Maze bbox:'
		for shape in shapes:
			print shape.bbox()

		print 'Combining.'
		dir = [Vector(-1, 0), Vector(0, -1), Vector(1, 0), Vector(0, 1)]
		head = choice(dir)
		comb = combine_shapes(shapes[0], shapes[1], head, ftex)
		print 'Done.'

		return comb

	scale = 0.05

	shape = make_shape()
	update = 1

	#def draw_seg(seg, lookat, color=(0, 255, 0)):
	#	p1 = seg.p1.mul(scale).subv(lookat).addv(sc).tupi()
	#	p2 = seg.p2.mul(scale).subv(lookat).addv(sc).tupi()
	#	pygame.draw.line(disp, color, p1, p2)

	lookat = Vector(0, 0)
	glEnable(GL_TEXTURE_2D)
	#glPolygonMode(GL_FRONT_AND_BACK, GL_LINE)
		
	glNewList(1, GL_COMPILE)
	for quad in shape.real_quads:
		quad.render()
	glEndList()

	k_state = {}

	while 1:
		for ev in pygame.event.get():
			if ev.type == KEYDOWN:
				if ev.key == K_ESCAPE:
					break
			if ev.type in [KEYUP, KEYDOWN]:
				k_state[ev.key] = ev.type == KEYDOWN

		if k_state.get(K_LEFT, 0):
			lookat = lookat.addi(0.05, 0)
		if k_state.get(K_RIGHT, 0):
			lookat = lookat.addi(-0.05, 0)
		if k_state.get(K_UP, 0):
			lookat = lookat.addi(0, 0.05)
		if k_state.get(K_DOWN, 0):
			lookat = lookat.addi(0, -0.05)
		if k_state.get(K_q, 0):
			scale += 0.01
		if k_state.get(K_a, 0):
			scale -= 0.01

		glClear(GL_COLOR_BUFFER_BIT)

		glLoadIdentity()
		glScalef(scale, scale, 1)
		glTranslatef(lookat.x, lookat.y, 0)

		glCallList(1)
	
		pygame.display.flip()

