"""Drivetrain math."""

from math2 import Vector, Matrix, Segment

class Object (object):
	def __init__(self):
		self.position = Vector(0, 0)
		self.rotation = Matrix.mident()
		self.lmomentum = Vector(0, 0)
		self.amomentum = 0.0
		self.friction_feedback = Vector(0, 0)

		self.reset()
	
	def move(self, v):
		self.position = self.position.addv(v)
	
	def add_force(self, v):
		self.lmomentum = self.lmomentum.addv(v)
	
	def apply_force(self, pos, force):
		radius = pos.subv(self.position)
		torque = radius.cross(force)
		self.add_force(force)
		self.add_torque(torque * 2)
		
	def add_torque(self, t):
		self.amomentum += t

	def add_friction(self, f):
		"""Friction is specified in body coordinates."""
		# Inverse of rotation.
		irot = self.rotation.transpose()
		om = irot.mulv(self.lmomentum)
		fr = self.rotation.mulv(Vector(-om.x * f.x, -om.y * f.y))
		self.friction_feedback = fr
		self.add_force(fr)

	def add_forward(self, amount):
		forward = Vector(0, 1).mul(amount)
		forward = self.rotation.mulv(forward)
		self.add_force(forward)

	def reset(self):
		self.force_acc = Vector(0, 0)

	def step(self, t):
		self.position = self.position.addv(self.lmomentum.mul(t))
		torqm = Matrix.mrot(self.amomentum * t)
		self.rotation = self.rotation.mulm(torqm)
		self.amomentum *= 0.5
		self.reset()

class TracksDrive (Object):
	def __init__(self, power, maxvel, turn):
		Object.__init__(self)
		self.power = float(power)
		self.maxvel = float(maxvel)
		self.turnspeed = float(turn)
		self.throttle = 0

	def step(self, t):
		if self.throttle:
			f_y = 0
		else:
			f_y = 0.25

		self.add_friction(Vector(0.99, f_y))
		Object.step(self, t)

	def set_throttle(self, amount):
		vel = self.lmomentum.length() / self.maxvel
		self.throttle = self.power * (1.0 - vel) * amount
		self.add_forward(self.throttle)
	
	def set_turn(self, amount):
		self.add_torque(self.turnspeed * amount)

	def reset(self):
		self.throttle = 0
		Object.reset(self)

class WheelDrive (Object):
	def __init__(self, power, maxvel, turn):
		Object.__init__(self)
		self.power = float(power)
		self.maxvel = float(maxvel)
		self.turnspeed = float(turn)
	
	def step(self, t):
		self.add_friction(Vector(0.1, 0.01))
		Object.step(self, t)

	def set_throttle(self, amount):
		vel = self.lmomentum.length() / self.maxvel
		power = self.power * amount * (1.0 - vel)
		self.add_forward(power)
	
	def set_turn(self, amount):
		# 0.7 is magic, because momentum.length() is never near maxvel
		vel = self.lmomentum.length() / self.maxvel
		prop = -16 * (vel - 0.2) ** 2 / (vel + 0.8) ** 2 + 1
		turn = prop * self.turnspeed * amount
		self.add_torque(turn)
		
	def reset(self):
		self.amomentum = 0.0
		Object.reset(self)

class Vehicle (object):
	def __init__(self, shape, drive, level):
		self.shape = shape
		self.drive = drive
		self.level = level
	
	def change_drive(self, new_drive):
		new_drive.position = self.drive.position
		new_drive.rotation = self.drive.rotation

	def step(self, t):
		drive = self.drive
		shape = self.shape

		drive.step(t)

		coll = self.level.collide(shape)
		shape.move(drive.position, drive.rotation)

		if coll:
			point, normal = coll
			assert(normal is not None)
		
			normseg = Segment(Vector(0, 0), normal)
			proj, pu = normseg.project(drive.lmomentum)

			drive.move(drive.lmomentum.mul(-1.1*t))
			drive.move(normal.mul(0.01))

			feedback = proj.mul(-1.5)
			drive.apply_force(point, feedback)

		shape.move(drive.position, drive.rotation)

if __name__ == '__main__':
	import pygame
	from pygame.locals import *
	from math2 import Shape
	from digger import Digger
	from backtracker import Backtracker
	from level import ShapeLevel
	from OpenGL.GL import *

	pygame.init()
	disp = pygame.display.set_mode((600, 600))
	clock = pygame.time.Clock()

	#teximg = pygame.image.load('fueltank.jpg')
	#tex = pygame.image.tostring(teximg, 'RGB')
	#
	#glBindTexture(GL_TEXTURE_2D, 1)
	#glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
	#glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
	#glTexImage2D(GL_TEXTURE_2D, 0, 3, teximg.get_width(), teximg.get_height(),
	#		0, GL_RGB, GL_UNSIGNED_BYTE, tex)
	#glEnable(GL_TEXTURE_2D)

	t = 0.1
	k_state = {}
	running = 1
	sc = Vector(300, 300)

	#level = Shape.rectangle(100, 100, 400, 400)
	maze = Backtracker(15, 15)
	maze.generate()
	maze_segs, maze_shapes = maze.build(0.4)
	level = ShapeLevel(Shape(maze_segs))

	shape = Shape.rectangle(-0.05, -0.1, 0.1, 0.2)
	#drive = WheelDrive(0.05, 5.0, 1.2)
	drive = TracksDrive(0.1, 1.5, 0.7)
	drive.move(Vector(0.5, 0.5))
	#drive.move(Vector(150, 150))
	vehicle = Vehicle(shape, drive, level)

	paused = 0
	scale = 70

	surf = pygame.image.load('data/bumblebee_bot_1.png')
	font = pygame.font.Font(pygame.font.get_default_font(), 12)

	def draw_seg(seg, color, lookat):
		p1 = seg.p1.subv(lookat).mul(scale).addv(sc).tupi()
		p2 = seg.p2.subv(lookat).mul(scale).addv(sc).tupi()
		pygame.draw.line(disp, color, p1, p2)

	def draw_poly(points, color, lookat):
		points = [p.subv(lookat).mul(scale).addv(sc).tupi()
				for p in points]
		pygame.draw.polygon(disp, color, points)

	#def draw_seg(seg, color):
	#	p1 = seg.p1
	#	p2 = seg.p2
	#	glColor3fv(color)
	#	glVertex2f(p1.x, p1.y)
	#	glVertex2f(p2.x, p2.y)
	#
	#def draw_wall(seg):
	#	wvec = seg.p2.subv(seg.p1)
	#	unit = wvec.normal().mul(0.2)

	#	p1 = seg.p1
	#	p2 = seg.p2
	#	p3 = p2.addv(unit.left())
	#	p4 = p1.addv(unit.left())

	#	glTexCoord2f(p1.x, p1.y)
	#	glVertex3f(p1.x, p1.y, 0)
	#	glTexCoord2f(p2.x, p2.y)
	#	glVertex3f(p2.x, p2.y, 0)
	#	glTexCoord2f(p3.x, p3.y)
	#	glVertex3f(p3.x, p3.y, 0)
	#	glTexCoord2f(p4.x, p4.y)
	#	glVertex3f(p4.x, p4.y, 0)

	#glNewList(1, GL_COMPILE)
	#glBegin(GL_QUADS)
	#for s in level.shape.real_segments:
	#	draw_wall(s)
	#glEnd()
	#glEndList()

	lv_colors = {
		'wall': (0, 128, 0),
		'floor': (0, 64, 0)
	}


	while running:
		for ev in pygame.event.get():
			if ev.type in [KEYUP, KEYDOWN]:
				k_state[ev.key] = ev.type == KEYDOWN
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				running = 0
			if ev.type == KEYDOWN and ev.key == K_SPACE:
				paused = 1 - paused

		if not paused:
			if k_state.get(K_LEFT, 0):
				vehicle.drive.set_turn(-1)
			if k_state.get(K_RIGHT, 0):
				vehicle.drive.set_turn(1)

			if k_state.get(K_UP, 0):
				vehicle.drive.set_throttle(1)
			if k_state.get(K_DOWN, 0):
				vehicle.drive.set_throttle(-1)

		vehicle.step(t)
		lookat = vehicle.drive.position
		#glClear(GL_COLOR_BUFFER_BIT)
		
		disp.fill(0)

		label = font.render('%d' % level.last_coll_segments, 1, 
				(255, 255, 255))
		disp.blit(label, (0, 0))

		for shape in maze_shapes:
			color = lv_colors[shape.texname]
			draw_poly(shape.points, color, lookat)

		for seg in vehicle.shape.real_segments:
			draw_seg(seg, (255, 255, 255), lookat)

		p1x, p1y, p2x, p2y = vehicle.shape.bbox()
		draw_seg(Segment(Vector(p1x, p1y), Vector(p2x, p2y)), 
				(255, 0, 0), lookat)


		#for seg in level.shape.real_segments:
		#	draw_seg(seg, (0, 255, 0), lookat)

		#glLoadIdentity()
		#glScalef(scale, -scale, 1)
		#glTranslatef(-lookat.x, -lookat.y, 0)

		#glDisable(GL_TEXTURE_2D)
		#glBegin(GL_LINES)
		#for s in vehicle.shape.real_segments:
		#	draw_seg(s, (255, 255, 255))
		#glEnd()
		#glEnable(GL_TEXTURE_2D)
		#glCallList(1)

		pygame.display.update()
		clock.tick(25)
