from random import uniform
from math import cos, sin, pi, sqrt

class Vector (object):
	def __init__(self, x, y):
		self.x = float(x)
		self.y = float(y)

	def addi(self, x, y):
		return Vector(self.x + x, self.y + y)
	
	def subi(self, x, y):
		return Vector(self.x - x, self.y - y)

	def addv(self, v):
		return Vector(self.x + v.x, self.y + v.y)
	
	def subv(self, v):
		return Vector(self.x - v.x, self.y - v.y)

	def mul(self, v):
		return Vector(self.x * v, self.y * v)

	def scale(self, v):
		return Vector(self.x * v.x, self.y * v.y)
	
	def dot(self, b):
		return self.x * b.x + self.y * b.y

	def cross(self, b):
		return self.x * b.y - self.y * b.x

	def tupi(self):
		return (int(self.x), int(self.y))

	def tuple(self):
		return (self.x, self.y)

	def abs(self):
		return Vector(abs(self.x), abs(self.y))

	def length(self):
		return sqrt(self.x ** 2 + self.y ** 2)
	
	def normal(self):
		return self.mul(1 / self.length())
	
	def left(self):
		return Vector(self.y, -self.x)

	def right(self):
		return Vector(-self.y, self.x)

	def randomize(self, v):
		return self.addi(uniform(-v, v), uniform(-v, v))

	def __repr__(self):
		return '<Vector(%.2f, %.2f)>' % (self.x, self.y)
	
	def __eq__(self, other):
		if not (hasattr(other, 'x') and hasattr(other, 'y')):
			return False
		return self.x == other.x and self.y == other.y
	
	def __ne__(self, other):
	 	return not (self == other)

	def __hash__(self):
		return hash((self.x, self.y))

class Matrix (object):
	def __init__(self, *args):
		self.m = map(float, args)
	
	def mulm(self, other):
		m = self.m
		n = other.m

		return Matrix(
			m[0] * n[0] + m[1] * n[2],
			m[0] * n[1] + m[1] * n[3],
			m[2] * n[0] + m[3] * n[2],
			m[2] * n[1] + m[3] * n[3])

	def mulv(self, v):
		m = self.m
		return Vector(
			v.x * m[0] + v.y * m[1],
			v.x * m[2] + v.y * m[3])

	def to_gl(self):
		"""Return array suitable for glMultMatrix."""
		m = self.m
		return [
			m[0], m[2], 0, 0,
			m[1], m[3], 0, 0,
			0,    0,    1, 0,
			0,    0,    0, 1]
	
	def __repr__(self):
		return '<Matrix(%s)>' % (', '.join(['%.2f' % v for v in self.m]))

	def transpose(self):
		m = self.m
		return Matrix(m[0], m[2], m[1], m[3])

	@classmethod
	def mrot(cls, angle):
		a = cos(angle)
		b = sin(angle)

		return Matrix(a, -b, b, a)

	@classmethod
	def mident(cls):
		return Matrix(1, 0, 0, 1)
	
class Segment (object):
	def __init__(self, p1, p2):
		self.p1 = p1
		self.p2 = p2

	def length(self):
		return self.p2.subv(self.p1).length()

	def __repr__(self):
		return '<Segment(%.2f, %.2f, %.2f, %.2f)>' % (
				self.p1.x, self.p1.y, self.p2.x, self.p2.y)

	def __eq__(self, other):
		if not isinstance(other, Segment):
			return False
		return (self.p1 == other.p1 and self.p2 == other.p2)

	def __ne__(self, other):
		if not isinstance(other, Segment):
			return True
		return (self.p1 != other.p1 or self.p2 != other.p2)

	def __hash__(self):
		return hash(self.p1.tuple() + self.p2.tuple())

	def tuple(self):
		return (self.p1.x, self.p1.y, self.p2.x, self.p2.y)

	def normal(self):
		return self.p2.subv(self.p1).normal().right()

	def intersect(self, b):
		x1 = self.p1.x
		y1 = self.p1.y
		x2 = self.p2.x
		y2 = self.p2.y
		x3 = b.p1.x
		y3 = b.p1.y
		x4 = b.p2.x
		y4 = b.p2.y
	
		d = (y4 - y3) * (x2 - x1) - (x4 - x3) * (y2 - y1)
	
		if d == 0:
			d2 = (x3 - x1) * (y2 - y1) - (y3 - y1) * (x2 - x1)
			
			if d2 == 0:
				# Coincident.
				if x1 == x2:
					t0 = (y3 - y1) / (y2 - y1)
					t1 = (y4 - y1) / (y2 - y1)
				else:
					t0 = (x3 - x1) / (x2 - x1)
					t1 = (x4 - x1) / (x2 - x1)
	
				if 0 <= t0 <= 1 or 0 <= t1 <= 1:
					t0 = min(1, max(0, t0))
					t1 = min(1, max(0, t1))
					
					x5 = x1 + t0 * (x2 - x1)
					y5 = y1 + t0 * (y2 - y1)
	
					x6 = x1 + t1 * (x2 - x1)
					y6 = y1 + t1 * (y2 - y1)
	
					if x5 == x6 and y5 == y6:
						return 0.0, Vector(x5, y5)
					else:
						return 0.0, Segment(Vector(x5, y5), Vector(x6, y6))
				else:
					l = [
						Segment(self.p1, b.p1),
						Segment(self.p1, b.p2),
						Segment(self.p2, b.p1),
						Segment(self.p2, b.p2)]
					l.sort(lambda a, b: cmp(a.length(), b.length()))
					return l[0].length(), l[0]
	
			else:
				# Parralel.
				return None, None
	
		else:
			ua = ((x4 - x3) * (y1 - y3) - (y4 - y3) * (x1 - x3)) / d
			ub = ((x2 - x1) * (y1 - y3) - (y2 - y1) * (x1 - x3)) / d
		
			if 0 <= ua <= 1 and 0 <= ub <= 1:
				# Intersect.
				px = x1 + ua * (x2 - x1)
				py = y1 + ua * (y2 - y1)
				return 0.0, Vector(px, py)
			
			else:
				# Do not intersect.
				p1, seg1 = b.distance(self.p1)
				p2, seg2 = b.distance(self.p2)

				if p1 < p2:
					return p1, seg1
				else:
					return p2, seg2
	
	def project(self, p):
		ab = self.p2.subv(self.p1)
		ap = p.subv(self.p1)
		x = self.p1.addv(ab.mul(ap.dot(ab) / ab.dot(ab)))
		if ab.x != 0.0:
			u = (x.x - self.p1.x) / ab.x
		else:
			u = (x.y - self.p1.y) / ab.y
	
		x = Vector(x.x, x.y)
		return x, u
	
	def distance(self, p):
		x, u = self.project(p)
	
		if 0 <= u <= 1:
			seg = Segment(p, x)
			return seg.length(), seg
		else:
			seg1 = Segment(p, self.p1)
			seg2 = Segment(p, self.p2)

			if seg1.length() < seg2.length():
				return seg1.length(), seg1
			else:
				return seg2.length(), seg2

	def transform(self, t, r):
		seg = Segment(r.mulv(self.p1).addv(t), r.mulv(self.p2).addv(t))
		if hasattr(self, 'quad'):
			seg.quad = self.quad
		return seg

	def middle(self):
		return self.p1.addv(self.p2.subv(self.p1).mul(0.5))
	
	def scale(self, v):
		seg = Segment(self.p1.mul(v), self.p2.mul(v))
		if hasattr(self, 'quad'):
			seg.quad = self.quad
		return seg

class Shape (object):
	def __init__(self, segments):
		self.segments = self.real_segments = segments
		self.position = Vector(0, 0)
		self.rotation = Matrix.mident()

	def move(self, position, rotation):
		self.position = position
		self.rotation = rotation
		self.update_real_segments()

	def update_real_segments(self):
		position = self.position
		rotation = self.rotation
		self.real_segments = [s.transform(position, rotation) 
				for s in self.segments]
	def bbox(self):
		minx = maxx = self.real_segments[0].p1.x
		miny = maxy = self.real_segments[0].p1.y
		for s in self.real_segments:
			for p in [s.p1, s.p2]:
				minx = min(minx, p.x)
				maxx = max(maxx, p.x)
				miny = min(miny, p.y)
				maxy = max(maxy, p.y)
		return (minx, miny, maxx, maxy)

	def center_and_size(self):
		ax, ay, bx, by = self.bbox()
		a = Vector(ax, ay)
		b = Vector(bx, by)
		size = b.subv(a)
		center = a.addv(size.mul(0.5))
		return center, size

	def remove(self,seg):
		self.segments.remove(seg)
		self.update_real_segments()

	def contacts(self, other):
		spos = self.position
		opos = other.position
		contacts = []

		for s1 in self.real_segments:
			for s2 in other.real_segments:
				dist, value = s1.intersect(s2)
				if dist == 0:
					contacts.append((value, s2))
		return contacts

	def collide(self, other):
		contacts = self.contacts(other)

		if not contacts:
			return None

		center = Vector(0, 0)
		normal = Vector(0, 0)
		count = 0
		for value, segment in contacts:
			if isinstance(value, Vector):
				center = center.addv(value)
				normal = normal.addv(segment.normal())
				count += 1
			elif isinstance(value, Segment):
				center = center.addv(value.p1).addv(value.p2)
				normal = normal.addv(segment.normal())
				count += 2

		center = center.mul(1. / count)
	
		# If we can't get normal (in case we crossed two
		# segments with whose normals were norm1 = -norm2), return
		# None and hope for the best.
		if normal.length():
			normal = normal.normal()
		else:
			normal = None
		
		return (center, normal)
	
	@classmethod
	def rectangle_segments(cls, x, y, w, h):
		p1 = Vector(x, y)
		p2 = Vector(x + w, y)
		p3 = Vector(x + w, y + h)
		p4 = Vector(x, y + h)
		return [Segment(p1, p2),
			Segment(p2, p3),
			Segment(p3, p4),
			Segment(p4, p1)]

	@classmethod
	def rectangle(cls, x, y, w, h):
		return cls(cls.rectangle_segments(x, y, w, h))

	def __repr__(self):
		return '<Shape(%s)>' % ', '.join(map(repr, self.segments))

	def __len__(self):
		return len(self.segments)

if __name__ == '__main__':
	import pygame, random
	pygame.init()
	disp = pygame.display.set_mode((300, 300))
	clock = pygame.time.Clock()

	mr = Matrix.mident()
	ms = Matrix.mrot(pi/50)
	ba = Vector(100, 0)
	sc = Vector(180, 150)

	while 1:
		mr = mr.mulm(ms)
		vr = mr.mulv(ba)
		vs = vr.addv(sc)

		segr = Segment(sc, vs)
		segv = Segment(Vector(100, 0), Vector(100, 300))

		dist, point = segr.intersect(segv)

		disp.fill(0)
		if isinstance(point, Vector):
			pygame.draw.circle(disp, (255, 0, 0), point.tupi(), 10, 0)
		elif isinstance(point, Segment) and dist < 75.0:
			p1 = point.p1.addi(random.randint(-5, 5), random.randint(-5, 5))
			p2 = point.p2.addi(random.randint(-5, 5), random.randint(-5, 5))
			pygame.draw.line(disp, (0, 255, 255), p1.tupi(), p2.tupi())

		norm = segr.normal().mul(15).addv(sc)

		pygame.draw.line(disp, (255, 255, 255), segv.p1.tupi(), segv.p2.tupi())
		pygame.draw.line(disp, (255, 255, 255), segr.p1.tupi(), segr.p2.tupi())
		pygame.draw.line(disp, (255, 0, 0), sc.tupi(), norm.tupi())
		pygame.display.update()

		clock.tick(25)

