import pygame
from math2 import Vector, Matrix, Segment, Shape
from OpenGL.GL import *
from data import get_data_file

class TextureManager(object):
	def __init__(self):
		self.textures = {}
	
	def __getitem__(self, key):
		try:
			return self.textures[key]
		except KeyError:
			tex = Texture(key)
			self.textures[key] = tex
			return tex

	def load(self):
		for tex in self.textures.values():
			tex.load()

class Texture(object):
	"""Texture(filename, wrap_s, wrap_t) -> Texture
	Create a texture for the renderer.
	"""
	
	def __init__(self, filename):
		self.filename = get_data_file(filename)
		self.tex = None
	
	def __repr__(self):
		return '<Texture(%s)>' % self.filename

	def load(self):
		tex_image = pygame.image.load(self.filename)
		self.w, self.h = tex_image.get_size()
		tex_data = pygame.image.tostring(tex_image, 'RGBA')

		self.tex = glGenTextures(1)

		self.select()
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR)
		glTexParameter(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR)
		glTexImage2D(GL_TEXTURE_2D, 0, 4, self.w, self.h, 0, GL_RGBA,
				GL_UNSIGNED_BYTE, tex_data)

	def select(self):
		glBindTexture(GL_TEXTURE_2D, self.tex)

class Quad(object):
	def __init__(self, texture, vertices, tcoords=None):
		self.texture = texture
		self.vertices = list(vertices)
		if tcoords:
			self.tcoords = list(tcoords)
		else:
			self.gen_tex_coords()

	def transform(self, trans, rot):
		new_v = [rot.mulv(v).addv(trans) for v in self.vertices]
		return type(self)(self.texture, new_v, self.tcoords)

	def scale(self, v):
		return type(self)(self.texture, 
			[p.mul(v) for p in self.vertices],
			self.tcoords)

	def render(self):
		self.texture.select()
		glBegin(GL_QUADS)
		for i in range(4):
			glTexCoord2fv(self.tcoords[i].tuple())
			glVertex2fv(self.vertices[i].tuple())
		glEnd()

class WallQuad(Quad):
	"""WallQuad(texture, vertices) -> WallQuad

	Automagically generate texture for walls.
	Wall textures repeat in s direction, and are stretched in t direction.
	"""
	def gen_tex_coords(self):
		v0, v1, v2, v3 = tuple(self.vertices)
		
		t1s = v1.subv(v0).length()

		# Basis vectors of texture plane.
		us = v1.subv(v0).normal()
		ut = us.left()

		seg = Segment(v2, v3)
		p0, u0 = seg.project(v0)
		p1, u1 = seg.project(v1)

		d2s = v2.subv(p1).length()
		if 0 <= u1 <= 1:
			t2s = t1s + d2s
		else:
			t2s = t1s - d2s

		d3s = v3.subv(p0).length()
		if 0 <= u0 <= 1:
			t3s = -d3s
		else:
			t3s = d3s
		
		t0 = Vector(0, 0)
		t1 = Vector(t1s, 0)
		t2 = Vector(t2s, 1)
		t3 = Vector(t3s, 1)

		self.tcoords = [t0, t1, t2, t3]

	def __repr__(self):
		return '<WallQuad(%s)>' % ', '.join(['%s' % v for v in self.vertices])

class FloorQuad(Quad):
	"""FloorQuad(texture, vertices, tcoords=None) -> FloorQuad

	Automagically generate texture coords for floors.
	Floor textures repeat in both directions.
	"""

	def gen_tex_coords(self):
		size = self.vertices[2].subv(self.vertices[0])
		
		#t0 = Vector(0, 0)
		#t1 = Vector(size.x, 0)
		#t2 = size
		#t3 = Vector(0, size.y)

		#self.tcoords = [t0, t1, t2, t3]
		self.tcoords = self.vertices[:]

class VisibleShape (Shape):
	def __init__(self, segments, quads):
		Shape.__init__(self, segments)
		self.quads = quads
		self.real_quads = quads
	
	def move(self, position, rotation):
		self.position = position
		self.rotation = rotation

		new_segs = []
		quad_trans = {}
		for seg in self.segments:
			new_seg = seg.transform(position, rotation)
			new_quad = seg.quad.transform(position, rotation)
			new_seg.quad = new_quad
			quad_trans[seg.quad] = new_quad
			new_segs.append(new_seg)

		new_quads = []
		for quad in self.quads:
			if quad in quad_trans:
				new_quad = quad_trans[quad]
			else:
				new_quad = quad.transform(position, rotation)
			new_quads.append(new_quad)

		self.real_segments = new_segs
		self.real_quads = new_quads

class Sprite(object):
	def __init__(self, w, h, textures):
		self.frames = textures

		w /= 2
		h /= 2

		self.mdl = [
			(1, 1, -w, -h),
			(0, 1, w, -h),
			(0, 0, w, h),
			(1, 0, -w, h)]

	def load(self):
		self.glist = glGenLists(1)
		glNewList(self.glist, GL_COMPILE)
		glPushAttrib(GL_CURRENT_BIT | GL_ENABLE_BIT)
		glEnable(GL_BLEND)
		glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA)
		glBegin(GL_QUADS)
		for s, t, x, y in self.mdl:
			glTexCoord2f(s, t)
			glVertex2f(x, y)
		glEnd()
		glPopAttrib()
		glEndList()

	def render(self, frame=0):
		self.frames[frame].select()
		glCallList(self.glist)

if __name__ == '__main__':
	import pygame
	from pygame.locals import *

	disp = pygame.display.set_mode((600, 600), DOUBLEBUF | OPENGL)
	clock = pygame.time.Clock()

	wtex = Texture('lab-wall.png')
	wtex.load()
	ftex = Texture('lab-floor2.png')
	ftex.load()

	floor = FloorQuad(ftex,
		[Vector(-1, -1), Vector(1, -1), Vector(1, 1), Vector(-1, 1)])

	wall1 = WallQuad(wtex,
		[Vector(-1, -1), Vector(1, -1), 
		Vector(1.1, -1.1), Vector(-1.1, -1.1)])
	wall2 = WallQuad(wtex,
		[Vector(-1, 1), Vector(1, 1),
		Vector(1.1, 1.1), Vector(-1.1, 1.1)])
	wall3 = WallQuad(wtex,
		[Vector(-1, -1), Vector(-1, 1),
		Vector(-1.1, 1.1), Vector(-1.1, -1.1)])
	wall4 = WallQuad(wtex,
		[Vector(1, -1), Vector(1, 1),
		Vector(1.1, 1.1), Vector(1.1, -1.1)])

	#glScalef(0.6, 0.6, 0.6)
	glEnable(GL_TEXTURE_2D)

	pos = Vector(0, 0)
	
	k_state = {}

	while 1:
		clock.tick(50)

		for ev in pygame.event.get():
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				break
			if ev.type in [KEYUP, KEYDOWN]:
				k_state[ev.key] = ev.type == KEYDOWN
		if k_state.get(K_LEFT, 0):
			pos = pos.addi(0.01, 0)
		if k_state.get(K_RIGHT, 0):
			pos = pos.addi(-0.01, 0)
		if k_state.get(K_UP, 0):
			pos = pos.addi(0, 0.01)
		if k_state.get(K_DOWN, 0):
			pos = pos.addi(0, -0.01)

		glLoadIdentity()
		glTranslatef(pos.x, pos.y, 0)

		glClear(GL_COLOR_BUFFER_BIT)
		floor.render()
		wall1.render()
		wall2.render()
		wall3.render()
		wall4.render()

		pygame.display.flip()

