import pygame, sys
from pygame.locals import *
from OpenGL.GL import *
from time import time

from digger import Digger
from backtracker import Backtracker
from drive import Vehicle, TracksDrive
from thelevel import TheLevelGenerator
from render import TextureManager, Sprite
from data import get_data_file
from math2 import Shape
from minimap import Minimap

class Scene (object):
	def __init__(self, game):
		self.game = game
	
	def initialize(self):
		pass
	def start(self):
		pass
	def stop(self):
		pass
	def step(self, t):
		pass
	def render(self):
		pass

class MusicManager (object):
	"""Multilayered BGM manager."""

	data_ambience = 'data/bgm/drones.ogg'
	data_melody = 'data/bgm/mix.ogg'
	data_tension = 'data/bgm/amen.ogg'
	
	def load(self):
		self.ambience = pygame.mixer.Sound(self.data_ambience)
		self.melody = pygame.mixer.Sound(self.data_melody)
		self.tension = pygame.mixer.Sound(self.data_tension)

		self.melody_volume = self.melody_volume_set = 0
		self.tension_volume = 0
	
	def start(self):
		# Start with melody and tension muted.
		self.melody.set_volume(self.melody_volume)
		self.tension.set_volume(self.melody_volume)

		self.chan_ambience = self.ambience.play(-1)
		self.chan_melody = self.melody.play(-1)
		self.chan_tension = self.tension.play(-1)
	
	def start_melody(self):
		self.melody_volume_set = 1.0
	
	def stop_melody(self):
		self.melody_volume_set = 0.0

	def step(self, t):
		delta = self.melody_volume_set - self.melody_volume
		if delta != 0:
			# Fade in over 4 seconds.
			self.melody_volume += t / 4.0 * delta / abs(delta)
		self.melody_volume = min(1.0, max(0.0, self.melody_volume))
		
		#self.ambience.set_volume(1. - self.melody_volume / 1.5)
		self.melody.set_volume(self.melody_volume)
		self.tension.set_volume(self.tension_volume)

class Player (object):
	def __init__(self, level, frames):
		self.drive = TracksDrive(0.1, 1.5, 0.7)
		self.shape = Shape.rectangle(-0.1, -0.1, 0.2, 0.2)
		self.sprite = Sprite(0.25, 0.25, frames)
		self.sprite.load()

		self.vehicle = Vehicle(self.shape, self.drive, level)
		self.health = 1.0
		self.frame = 0.0

		self.hit_sound = pygame.mixer.Sound(get_data_file('explosion.wav'))

	def take_hit(self):
		self.hit_sound.play()
		self.health -= 0.05

	def step(self, t):
		self.vehicle.step(t)
		self.frame += self.drive.lmomentum.length()
		if self.frame >= 2:
			self.frame -= 2.

	def render(self):
		glPushMatrix()
		glTranslatef(self.drive.position.x, self.drive.position.y, 0)
		glMultMatrixf(self.drive.rotation.to_gl())
		self.sprite.render(int(self.frame))

		glLoadIdentity()
		glPushAttrib(GL_ENABLE_BIT | GL_CURRENT_BIT)
		glDisable(GL_TEXTURE_2D)
		glColor3f(1, 1, 1)
		glBegin(GL_QUADS)
		glVertex2f(-0.4, -0.4)
		glVertex2f(-0.4, -0.42)
		glVertex2f(0.8 * self.health, -0.42)
		glVertex2f(0.8 * self.health, -0.4)
		glEnd()
		glPopAttrib()
		glPopMatrix()

class TitleScene (Scene):
	def initialize(self):
		image = pygame.image.load(get_data_file('title.png'))
		self.tsize = image.get_size()
		self.title = pygame.image.tostring(image, 'RGB', 1)
		self.areas = [
			('game', 64, 390, 315, 535),
			('exit', 75, 565, 285, 640)
		]

	def start(self):
		glLoadIdentity()
		glTranslatef(-0.5, -0.5, 0)

	def step(self, t):
		go = None
		for ev in pygame.event.get():
			if ev.type == MOUSEBUTTONDOWN:
				go = self.check_click(ev.pos)
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				return 'exit'
		return go

	def check_click(self, p):
		x, y = p
		for next, x1, y1, x2, y2 in self.areas:
			if x1 < x < x2 and y1 < y < y2:
				return next
		return None

	def render(self):
		glClear(GL_COLOR_BUFFER_BIT)
		w, h = self.tsize
		glDrawPixels(w, h, GL_RGB, GL_UNSIGNED_BYTE, self.title)

class GameScene (Scene):
	player_frames = [ 'bbb-tracks.0.png', 'bbb-tracks.1.png' ]
	
	def __init__(self, game):
		Scene.__init__(self, game)
		self.initialized = 0
		self.k_state = {}
		self.show_minimap = 0
		self.sounds = {}

	def initialize(self):
		self.init_world()
		self.init_player()
		self.game.textures.load()
		self.level.compile()
		self.minimap = Minimap(self.level)
		self.minimap.compile()
		self.initialized = 1

	def init_world(self):
		tlg = TheLevelGenerator(self.game.textures)
		self.level = tlg.create_the_level(self)

	def init_player(self):
		tt = [self.game.textures[t] for t in self.player_frames]
		self.player = Player(self.level, tt)

	def get_player_loc(self):
		return self.player.drive.position

	def get_player_bbox(self):
		return self.player.shape.bbox()

	def get_sound(self, name):
		try:
			return self.sounds[name]
		except KeyError:
			sound = pygame.mixer.Sound(get_data_file(name))
			self.sounds[name] = sound
			return sound

	def start(self):
		if not self.initialized:
			self.initialize()
		self.game.bgm.start_melody()
		glEnable(GL_TEXTURE_2D)
		glEnable(GL_LINE_SMOOTH)

	def step(self, t):
		self.level.step(t, self.player)

		if self.player.health < 0:
			return 'gameover'

		k_state = self.k_state
		for ev in pygame.event.get():
			if ev.type in [KEYUP, KEYDOWN]:
				k_state[ev.key] = ev.type == KEYDOWN
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				return 'title'
			if ev.type == KEYDOWN and ev.key == K_TAB:
				self.show_minimap = 1 - self.show_minimap

		vehicle = self.player.vehicle

		if k_state.get(K_LEFT, 0):
			vehicle.drive.set_turn(1)
		if k_state.get(K_RIGHT, 0):
			vehicle.drive.set_turn(-1)

		if k_state.get(K_UP, 0):
			vehicle.drive.set_throttle(1)
		if k_state.get(K_DOWN, 0):
			vehicle.drive.set_throttle(-1)

		self.player.step(t)
	
	def stop(self):
		self.game.bgm.stop_melody()
		glDisable(GL_TEXTURE_2D)

	def render(self):
		drive = self.player.drive
		glClear(GL_COLOR_BUFFER_BIT)
		glLoadIdentity()
		glScalef(0.2, 0.2, 0.2)
		glMultMatrixf(drive.rotation.transpose().to_gl())
		glTranslatef(-drive.position.x, -drive.position.y, 0)
		self.level.render()
		self.player.render()

		if self.show_minimap:
			glLoadIdentity()
			glScalef(0.01, 0.01, 0.01)
			glMultMatrixf(drive.rotation.transpose().to_gl())
			glTranslatef(-drive.position.x, -drive.position.y, 0)
			glColor3f(0, 1, 0)
			self.minimap.render()
			glColor3f(0, 0.5, 0)
		else:
			glColor3f(1, 1, 1)

class GameOverScene(Scene):
	def start(self):
		self.game.show_image('gameover.png')
		self.duration = 2
	
	def step(self, t):
		self.duration -= t

		if self.duration < 0:
			return 'title'


class ExitScene(Scene):
	def start(self):
		sys.exit(0)

class Game (object):
	"""Handle everything"""
	def __init__(self):
		# Called at game start.
		self.running = 0
		self.bgm = MusicManager()
		self.textures = TextureManager()

	def init_scenes(self):
		self.scenes = {
			'title': TitleScene(self),
			'game': GameScene(self),
			'gameover': GameOverScene(self),
			'exit': ExitScene(self)
		}
		self.current_scene = self.scenes['title']

	def init(self):
		pygame.init()
		pygame.display.set_mode((1024, 768), OPENGL | DOUBLEBUF)
		glMatrixMode(GL_PROJECTION)
		glLoadIdentity()
		glOrtho(-.4, .4, -.3, .3, -1, 1)
		glMatrixMode(GL_MODELVIEW)

		self.show_image('loading.png')
		self.bgm.load()
		self.init_scenes()
		for scene in self.scenes.values():
			scene.initialize()

	def show_image(self, image):
		path = get_data_file(image)
		image = pygame.image.load(path)
		string = pygame.image.tostring(image, 'RGB', 1)
		w, h = image.get_size()
		glLoadIdentity()
		glTranslate(-0.5, -0.5, 0)
		glDrawPixels(w, h, GL_RGB, GL_UNSIGNED_BYTE, string)
		pygame.display.flip()
	
	def run(self):
		"""Main loop."""
		clock = pygame.time.Clock()
		
		self.running = 1	
		last_time = time()

		self.bgm.start()
		self.current_scene.start()

		while self.running:
			clock.tick(30)

			if pygame.event.get(QUIT):
				sys.exit(0)

			current_time = time()
			elapsed_time = current_time - last_time
			last_time = current_time

			self.bgm.step(elapsed_time)
			switch = self.current_scene.step(elapsed_time)
			if switch:
				self.current_scene.stop()
				self.current_scene = self.scenes[switch]
				self.current_scene.start()
				continue

			self.current_scene.render()
			pygame.display.flip()

if __name__ == '__main__':
	game = Game()
	game.init()
	game.run()
