"""Classes of ingame items."""

from math import acos, pi
from math2 import Vector, Matrix
from entity import Entity
from OpenGL.GL import *

class WallTurret (Entity):
	def __init__(self, engine, loc, heading, sprite):
		Entity.__init__(self, engine, loc, heading)

		self.sprite = sprite
		self.aim = heading
		self.range = 2.0
		self.angle = 0.0
		self.aim_speed = pi / 8
		self.min_angle = -2 * pi / 5
		self.max_angle = 2 * pi / 5
		self.charge = 0.
		self.max_charge = 1.
		self.beam = None
		self.fire_sound = engine.get_sound('laser-shot.wav')
		self.fire_sound.set_volume(0.25)

		if heading == Vector(1, 0):
			angle = pi/2
		elif heading == Vector(0, 1):
			angle = pi
		elif heading == Vector(-1, 0):
			angle = 3*pi/2
		elif heading == Vector(0, -1):
			angle = 0

		self.rotation = Matrix.mrot(angle).to_gl()

	def step(self, t):
		player_loc = self.engine.get_player_loc()
		to_player = player_loc.subv(self.loc)
		
		far = to_player.length()

		self.turn = 0
		self.beam = None

		if far > self.range or to_player.dot(self.heading) < 0:
			self.charge = 0
			return

		self.charge = min(self.max_charge, self.charge + t)

		turn = acos(to_player.normal().dot(self.aim.left())) - pi/2
		max_turn = self.aim_speed * t
		turn = min(max_turn, max(-max_turn, turn))
		self.angle += turn
		self.angle = min(self.max_angle, max(self.min_angle, self.angle))
		mrot = Matrix.mrot(self.angle)
		self.aim = mrot.mulv(self.heading)

		self.turn = turn

		if self.charge >= self.max_charge:
			self.fire_sound.play()
			self.beam = self.aim.mul(self.range).addv(self.loc)
			self.charge = 0
		else:
			self.beam = None

	def render(self):
		glPushMatrix()
		glTranslatef(self.loc.x, self.loc.y, 0)
		glMultMatrixf(self.rotation)
		glTranslatef(0, -0.025, 0)
		self.sprite.render()
		glPopMatrix()

	def bbox(self):
		r = Vector(self.range, self.range).scale(self.heading)
		p1 = self.loc.addv(r.left()).addv(r)
		p2 = self.loc.addv(r.right()).addv(r)

		return (p1.x, p1.y, p2.x, p2.y)

if __name__ == '__main__':
	import pygame
	from pygame.locals import *
	from math2 import Vector
	from time import time

	pygame.init()
	disp = pygame.display.set_mode((600, 600))
	clock = pygame.time.Clock()
	mpos = Vector(0, 0)

	class EngineDummy (object):
		def set_player_loc(self, loc):
			self.loc = loc
		def get_player_loc(self):
			return self.loc
		def get_sound(self, sound):
			return pygame.mixer.Sound(sound)

	engine = EngineDummy()

	turrets = [
		WallTurret(engine, Vector(0, 300), Vector(1, 0)),
		WallTurret(engine, Vector(300, 0), Vector(0, 1)),
		WallTurret(engine, Vector(599, 300), Vector(-1, 0)),
		WallTurret(engine, Vector(300, 599), Vector(0, -1)),
		WallTurret(engine, Vector(300, 250), Vector(0, -1)),
		WallTurret(engine, Vector(300, 350), Vector(0, 1))
	]

	for t in turrets:
		t.range = 300

	font = pygame.font.Font(pygame.font.get_default_font(), 12)
	white = (255, 255, 255)
	red = (255, 0, 0)
	blue = (128, 255, 255)

	last_time = time()

	while 1:
		clock.tick(25)
		current_time = time()
		elapsed = current_time - last_time
		last_time = current_time

		for ev in pygame.event.get():
			if ev.type == KEYDOWN and ev.key == K_ESCAPE:
				break
			if ev.type == MOUSEMOTION:
				mpos = Vector(*ev.pos)

		disp.fill(0)
		engine.set_player_loc(mpos)
		for turret in turrets:
			turret.think(elapsed)
			
			p1 = turret.loc.tupi()
			p2 = turret.loc.addv(turret.aim.mul(20)).tupi()

			if turret.charge:
				pygame.draw.circle(disp, blue, p2, int(10 * turret.charge))

			pygame.draw.aaline(disp, white, p1, p2, 1)

			label = font.render('%.2f %.2f' % (
				turret.angle, turret.turn), 1, white)
			fp = turret.loc.addv(turret.heading.mul(50)).tupi()
			disp.blit(label, fp)


			if turret.beam:
				p3 = turret.loc.addv(turret.beam).tupi()
				pygame.draw.aaline(disp, red, p2, p3, 2)

		pygame.display.update()
